/*
 * Copyright (c) 2005-2010 KOM – Multimedia Communications Lab
 *
 * This file is part of Simonstrator.KOM.
 * 
 * Simonstrator.KOM is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 * 
 * PeerfactSim.KOM is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with PeerfactSim.KOM.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package de.tudarmstadt.maki.simonstrator.api.component.sensor.route;

import de.tudarmstadt.maki.simonstrator.api.component.sensor.SensorComponent;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.route.Route.RouteSegment;

/**
 * A sensor that provides access to the chosen route of a component. This is
 * only useful within mobility models that define routing to some extend (e.g.,
 * SUMO or OSM-based mobility models).
 * 
 * 
 * @author Tobias Meuser, Bjoern Richerzhagen
 *
 */
public interface RouteSensor extends SensorComponent {

	/**
	 * Returns the current route of the component.
	 * 
	 * @return
	 */
	public Route getRoute();

	/**
	 * Adds a listener for route updates
	 * 
	 * @param listener
	 */
	public void addRouteListener(RouteListener listener);

	/**
	 * Removes a listener for route updates
	 * 
	 * @param listener
	 */
	public void removeRouteListener(RouteListener listener);

	/**
	 * Adds a listener for segment updates along routes
	 * 
	 * @param listener
	 */
	public void addRouteSegmentListener(RouteSegmentListener listener);

	/**
	 * Removes a listener for segment updates along routes
	 * 
	 * @param listener
	 */
	public void removeRouteSegmentListener(RouteSegmentListener listener);

	/**
	 * A listener to be informed of each segment change along a route. This is
	 * realized as a separate listener, as the required computations should only
	 * be performed in case anyone is interested at all :)
	 * 
	 * @author Bjoern Richerzhagen
	 *
	 */
	public interface RouteSegmentListener {

		/**
		 * A route consists of several segments (basically street segments)
		 * along which the component moves towards the target. Whenever a
		 * component updates its current segment, this method is invoked.
		 * 
		 * @param route
		 * @param leftSegment
		 * @param enteredSegment
		 */
		public void onChangedRouteSegment(RouteSensor sensor, Route route, RouteSegment leftSegment,
				RouteSegment enteredSegment);

	}

	/**
	 * A listener for route updates and movement along the route.
	 * 
	 * @author Bjoern Richerzhagen
	 *
	 */
	public interface RouteListener {

		/**
		 * Triggered whenever the component reached the end of the current
		 * route.
		 * 
		 * @param sensor
		 * @param route
		 */
		public void onReachedDestination(RouteSensor sensor, Route route);

		/**
		 * Triggered whenever the route of the current component changes (either
		 * due to a new target being chosen or due to updates along the route).
		 * 
		 * @param sensor
		 * @param oldRoute
		 *            (may be null!)
		 * @param newRoute
		 */
		public void onChangedRoute(RouteSensor sensor, Route oldRoute, Route newRoute);

	}

}
